import random

import networkx as nx
from networkx.algorithms import tree


def predecessors(h, root):
    return {a: b for a, b in nx.bfs_predecessors(h, root)}


def random_spanning_tree(graph, pop_col):
    for edge in graph.edges:
        graph.edges[edge]["weight"] = random.random()

    edges = tree.maximum_spanning_edges(graph, algorithm="kruskal", data=False)
    spanning_tree = nx.Graph(edges)
    for node in graph:
        spanning_tree.nodes[node][pop_col] = graph.nodes[node][pop_col]
    return spanning_tree


def tree_part2(
    graph,
    pop_col,
    pop_target,
    epsilon,
    node_repeats,
    restarts=0,
    spanning_tree=None,
    choice=random.choice,
):
    """This function finds a balanced 2 partition of a graph by drawing a
    spanning tree and finding an edge to cut that leaves at most an epsilon
    imbalance between the populations of the parts. If a root fails, new roots
    are tried until node_repeats in which case a new tree is drawn.

    Builds up a connected subgraph with a connected complement whose population
    is ``epsilon * pop_target`` away from ``pop_target``.

    Returns a subset of nodes of ``graph`` (whose induced subgraph is connected).
    The other part of the partition is the complement of this subset.

    :param graph: The graph to partition
    :param pop_col: The node attribute holding the population of each node
    :param pop_target: The target population for the returned subset of nodes
    :param epsilon: The allowable deviation from  ``pop_target`` (as a percentage of
        ``pop_target``) for the subgraph's population
    :param node_repeats: A parameter for the algorithm: how many different choices
        of root to use before drawing a new spanning tree.
    :param restarts: Number of iterations (used when the algorithm chooses a new root)
    :param spanning_tree: The spanning tree for the algorithm to use (used when the
        algorithm chooses a new root and for testing)
    :param choice: :func:`random.choice`. Can be substituted for testing.
    """

    if spanning_tree is None:
        spanning_tree = random_spanning_tree(graph, pop_col)

    h = spanning_tree.copy()

    # this used to be greater than 2 but failed on small grids:(
    root = choice([x for x in spanning_tree.nodes if spanning_tree.degree(x) > 1])

    # BFS predecessors for iteratively contracting leaves
    pred = predecessors(h, root)

    # As we contract leaves, we keep track of which nodes merged together in
    # this dictionary:
    subsets = {x: {x} for x in graph}

    while len(h) > 1:
        leaves = [x for x in h if h.degree(x) == 1 and x != root]

        for leaf in leaves:
            if abs(h.nodes[leaf][pop_col] - pop_target) < epsilon * pop_target:
                return subsets[leaf]
            # Contract the leaf:
            parent = pred[leaf]
            h.nodes[parent][pop_col] += h.nodes[leaf][pop_col]
            subsets[parent] |= subsets[leaf]
            h.remove_node(leaf)

    if restarts < node_repeats:
        # Try again with new root, same tree
        return tree_part2(
            graph,
            pop_col,
            pop_target,
            epsilon,
            node_repeats,
            restarts + 1,
            spanning_tree,
        )
    else:
        # If restarts == node_repeats, start over completely with a new tree
        return tree_part2(graph, pop_col, pop_target, epsilon, node_repeats)



from collections import defaultdict


def level_sets(mapping: dict, container=set):
    """Inverts a dictionary. ``{key: value}`` becomes
    ``{value: <container of keys that map to value>}``."""
    sets = defaultdict(container)
    for source, target in mapping.items():
        sets[target].add(source)
    return sets



import collections
import math

from .flows import on_flow
from .cut_edges import on_edge_flow


def compute_polsby_popper(area, perimeter):
    try:
        return 4 * math.pi * area / perimeter**2
    except ZeroDivisionError:
        return math.nan


def polsby_popper(partition):
    return {part: compute_polsby_popper(partition['area'][part], partition['perimeter'][part])
            for part in partition.parts}


def boundary_nodes(partition, alias='boundary_nodes'):
    if partition.parent:
        return partition.parent[alias]
    return {node for node in partition.graph.nodes if partition.graph.nodes[node]['boundary_node']}


def initialize_exterior_boundaries_as_a_set(partition):
    part_boundaries = collections.defaultdict(set)
    for node in partition['boundary_nodes']:
        part_boundaries[partition.assignment[node]].add(node)
    return part_boundaries


@on_flow(initialize_exterior_boundaries_as_a_set, alias='exterior_boundaries_as_a_set')
def exterior_boundaries_as_a_set(partition, previous, inflow, outflow):
    graph_boundary = partition['boundary_nodes']
    return (previous | (inflow & graph_boundary)) - outflow


def initialize_exterior_boundaries(partition):
    graph_boundary = partition['boundary_nodes']
    boundaries = collections.defaultdict(lambda: 0)
    for node in graph_boundary:
        part = partition.assignment[node]
        boundaries[part] += partition.graph.nodes[node]['boundary_perim']
    return boundaries


@on_flow(initialize_exterior_boundaries, alias='exterior_boundaries')
def exterior_boundaries(partition, previous, inflow, outflow):
    graph_boundary = partition['boundary_nodes']
    added_perimeter = sum(partition.graph.nodes[node]['boundary_perim']
                          for node in inflow & graph_boundary)
    removed_perimeter = sum(partition.graph.nodes[node]['boundary_perim']
                            for node in outflow & graph_boundary)
    return previous + added_perimeter - removed_perimeter


def initialize_interior_boundaries(partition):
    return {part: sum(partition.graph.edges[edge]['shared_perim']
                      for edge in partition['cut_edges_by_part'][part])
            for part in partition.parts}


@on_edge_flow(initialize_interior_boundaries, alias='interior_boundaries')
def interior_boundaries(partition, previous, new_edges, old_edges):
    added_perimeter = sum(partition.graph.edges[edge]['shared_perim'] for edge in new_edges)
    removed_perimeter = sum(partition.graph.edges[edge]['shared_perim'] for edge in old_edges)
    return previous + added_perimeter - removed_perimeter


def flips(partition):
    return partition.flips


def perimeter_of_part(partition, part):
    """
    Totals up the perimeter of the part in the partition.
    Requires that 'boundary_perim' be a node attribute, 'shared_perim' be an edge
    attribute, 'cut_edges' be an updater, and 'exterior_boundaries' be an updater.
    """
    exterior_perimeter = partition['exterior_boundaries'][part]
    interior_perimeter = partition['interior_boundaries'][part]

    return exterior_perimeter + interior_perimeter


def perimeter(partition):
    return {part: perimeter_of_part(partition, part) for part in partition.parts}



import random


def mpaccept(partition):
    # if partition.parent is None:
    # return True
    val = 1
    # if (partition["POPB%"][3] > .5) and (partition["POPB%"][4] > .5):
    # return True
    if (partition["POPB%"][1] < .44) and partition["POPB%"][1] < partition.parent[
        "POPB%"
    ][1]:
        val = 0
    if (partition["POPB%"][3] < .44) and partition["POPB%"][3] < partition.parent[
        "POPB%"
    ][3]:
        val = 0
    if (partition["POPB%"][4] < .44) and partition["POPB%"][4] < partition.parent[
        "POPB%"
    ][4]:
        val = 0
    return (val + .05) > random.random()


def always_accept(partition):
    return True


def cut_edge_accept(partition):
    """Always accepts the flip if the number of cut_edges increases.
    Otherwise, uses the Metropolis criterion to decide.

    :param partition: The current partition to accept a flip from.
    :return: True if accepted, False to remain in place

    """
    bound = 1

    if partition.parent is not None:
        bound = min(1, len(partition.parent["cut_edges"]) / len(partition["cut_edges"]))

    return random.random() < bound


# def metagraph_accept(partition):
#     """Always accepts the flip if the metagraph degree increases.
#     Otherwise, uses the Metropolis criterion to decide.

#     :param partition: The current partition to accept a flip from.
#     :return: True if accepted, False to remain in place

#     """
#     # Doesn't work currently
#     bound = 1

#     if partition.parent is not None:
#         bound = min(1, len(partition.parent["metagraph_degree"])
#             / len(partition["metagraph_degree"]),
#         )

#     return random.random() < bound



import numpy


def mean_median(election_results):
    """
    Computes the Mean-Median score for the given ElectionResults.
    A positive value indicates an advantage for the first party listed
    in the Election's parties_to_columns dictionary.
    """
    first_party = election_results.election.parties[0]
    data = election_results.percents(first_party)

    return numpy.median(data) - numpy.mean(data)


def mean_thirdian(election_results):
    """
    Computes the Mean-Median score for the given ElectionResults.
    A positive value indicates an advantage for the first party listed
    in the Election's parties_to_columns dictionary.

    The motivation for this score is that the minority party in many
    states struggles to win even a third of the seats.
    """
    first_party = election_results.election.parties[0]
    data = election_results.percents(first_party)

    thirdian_index = round(len(data) / 3)
    thirdian = sorted(data)[thirdian_index]

    return thirdian - numpy.mean(data)


def efficiency_gap(results):
    """
    Computes the efficiency gap for the given ElectionResults.
    A positive value indicates an advantage for the first party listed
    in the Election's parties_to_columns dictionary.
    """
    party1, party2 = [results.counts(party) for party in results.election.parties]
    wasted_votes_by_part = map(wasted_votes, party1, party2)
    total_votes = results.total_votes()
    numerator = sum(waste2 - waste1 for waste1, waste2 in wasted_votes_by_part)
    return numerator / total_votes


def wasted_votes(party1_votes, party2_votes):
    """
    Computes the wasted votes for each party in the given race.
    :party1_votes: the number of votes party1 received in the race
    :party2_votes: the number of votes party2 received in the race
    """
    total_votes = party1_votes + party2_votes
    if party1_votes > party2_votes:
        party1_waste = party1_votes - total_votes / 2
        party2_waste = party2_votes
    else:
        party2_waste = party2_votes - total_votes / 2
        party1_waste = party1_votes
    return party1_waste, party2_waste



import random

from .tree_methods import tree_part2


def recom(partition, pop_col, pop_target, epsilon, node_repeats):
    """ReCom proposal.

    Description from MGGG's 2018 Virginia House of Delegates report:
    At each step, we (uniformly) randomly select a pair of adjacent districts and
    merge all of their blocks in to a single unit. Then, we generate a spanning tree
    for the blocks of the merged unit with the Kruskal/Karger algorithm. Finally,
    we cut an edge of the tree at random, checking that this separates the region
    into two new districts that are population balanced.

    Example usage::

        from functools import partial
        from gerrychain import MarkovChain
        from gerrychain.tree_proposals import recom

        # ...define constraints, accept, partition, total_steps here...

        # Ideal population:
        pop_target = sum(partition["population"].values()) / len(partition)

        proposal = partial(
            recom, pop_col="POP10", pop_target=pop_target, epsilon=.05, node_repeats=10
        )

        chain = MarkovChain(proposal, constraints, accept, partition, total_steps)

    """
    edge = random.choice(tuple(partition["cut_edges"]))
    parts_to_merge = (partition.assignment[edge[0]], partition.assignment[edge[1]])

    subgraph = partition.graph.subgraph(
        partition.parts[parts_to_merge[0]] | partition.parts[parts_to_merge[1]]
    )

    flips = recursive_tree_part(
        subgraph,
        parts_to_merge,
        pop_col=pop_col,
        pop_target=pop_target,
        epsilon=epsilon,
        node_repeats=node_repeats,
    )

    return flips


def recursive_tree_part(graph, parts, pop_target, pop_col, epsilon, node_repeats=20):
    """helper function for merge walk partitioning.
    Uses :func:`~gerrychain.tree_methods.tree_part2` recursively to partition a tree into
    ``len(parts)`` parts of population ``pop_target`` (within ``epsilon``).

    :param graph: The graph
    :param parts: Iterable of part labels (like ``[0,1,2]`` or ``range(4)``
    :param pop_target: Target population for each part of the partition
    :param pop_col: Node attribute key holding population data
    :param epsilon: How far (as a percentage of ``pop_target``) from ``pop_target`` the parts
        of the partition can be
    :param node_repeats: Parameter for :func:`~gerrychain.tree_methods.tree_part2` to use.
    :return: New assignments for the nodes of ``graph``.
    :rtype: dict
    """
    flips = {}
    remaining_nodes = set(graph.nodes)

    for part in parts[:-1]:
        nodes = tree_part2(
            graph.subgraph(remaining_nodes), pop_col, pop_target, epsilon, node_repeats
        )

        for node in nodes:
            flips[node] = part
        # update pop_target?

        remaining_nodes -= nodes

    # All of the remaining nodes go in the last part
    for node in remaining_nodes:
        flips[node] = parts[-1]

    return flips



class MetagraphDegree:

    """
    Updater that computes the metagraph degree of a proposed partition; i.e.,
    the number of possible single flips to make to turn into another partition.

    """

    def __init__(self, validator, alias):
        """
        :validator: :class:`~gerrychain.validty.Validator` instance.

        """
        self.validator = validator
        self.alias = alias

    def __call__(self, partition):
        total_available_flips = 2 * len(partition['cut_edges'])

        total_valid_flips = sum(self.num_valid_flips(edge, partition)
                                for edge in partition['cut_edges'])

        return {'total': total_available_flips, 'valid': total_valid_flips}

    def num_valid_flips(self, edge, partition):
        """Determine number of valid flips for the given edge (0, 1, or 2)."""
        flip = {edge[0]: partition.assignment[edge[1]]}
        reverse_flip = {edge[1]: partition.assignment[edge[0]]}

        # XXX: This is really awful.
        del partition.updaters[self.alias]
        flip_valid = self.validator(partition.merge(flip))
        reverse_valid = self.validator(partition.merge(reverse_flip))
        partition.updaters[self.alias] = self

        return flip_valid + reverse_valid



import json
import warnings

import geopandas as gp
import networkx
from networkx.readwrite import json_graph
from shapely.ops import cascaded_union
from shapely.prepared import prep

from .adjacency import neighbors
from .geo import reprojected


class Graph(networkx.Graph):
    """Represents a graph to be partitioned. It is based on :class:`networkx.Graph`.

    We have added some classmethods to help construct graphs from shapefiles, and
    to save and load graphs as JSON files.
    """

    @classmethod
    def from_json(cls, json_file):
        """Load a graph from a JSON file in the NetworkX json_graph format.
        :param json_file: Path to JSON file.
        :return: Graph
        """
        with open(json_file) as f:
            data = json.load(f)
        g = json_graph.adjacency_graph(data)
        graph = cls(g)
        graph.issue_warnings()
        return graph

    def to_json(self, json_file, *, include_geometries_as_geojson=False):
        """Save a graph to a JSON file in the NetworkX json_graph format.
        :param json_file: Path to target JSON file.
        :param bool include_geometry_as_geojson: (optional) Whether to include any
            :mod:`shapely` geometry objects encountered in the graph's node attributes
            as GeoJSON. The default (``False``) behavior is to remove all geometry
            objects because they are not serializable. Including the GeoJSON will result
            in a much larger JSON file.
        """
        data = json_graph.adjacency_data(self)

        if include_geometries_as_geojson:
            convert_geometries_to_geojson(data)
        else:
            remove_geometries(data)

        with open(json_file, "w") as f:
            json.dump(data, f)

    @classmethod
    def from_file(cls, filename, adjacency="rook", cols_to_add=None, reproject=True):
        """Create a :class:`Graph` from a shapefile (or GeoPackage, or GeoJSON, or
        any other library that :mod:`geopandas` can read. See :meth:`from_geodataframe`
        for more details.

        :param cols_to_add: (optional) The names of the columns that you want to
            add to the graph as node attributes. By default, all columns are added.
        """
        df = gp.read_file(filename)
        graph = cls.from_geodataframe(df, adjacency, reproject)
        graph.add_data(df, columns=cols_to_add)
        return graph

    @classmethod
    def from_geodataframe(cls, dataframe, adjacency="rook", reproject=True):
        """Creates the adjacency :class:`Graph` of geometries described by `dataframe`.
        The areas of the polygons are included as node attributes (with key `area`).
        The shared perimeter of neighboring polygons are included as edge attributes
        (with key `shared_perim`).
        Nodes corresponding to polygons on the boundary of the union of all the geometries
        (e.g., the state, if your dataframe describes VTDs) have a `boundary_node` attribute
        (set to `True`) and a `boundary_perim` attribute with the length of this "exterior"
        boundary.

        By default, areas and lengths are computed in a UTM projection suitable for the
        geometries. This prevents the bizarro area and perimeter values that show up when
        you accidentally do computations in Longitude-Latitude coordinates. If the user
        specifies `reproject=False`, then the areas and lengths will be computed in the
        GeoDataFrame's current coordinate reference system. This option is for users who
        have a preferred CRS they would like to use.

        :param dataframe: :class:`geopandas.GeoDataFrame`
        :param adjacency: (optional) The adjacency type to use ("rook" or "queen").
            Default is "rook".
        :return: The adjacency graph of the geometries from `dataframe`.
        :rtype: :class:`Graph`
        """
        # Project the dataframe to an appropriate UTM projection unless
        # explicitly told not to.
        if reproject:
            df = reprojected(dataframe)
        else:
            df = dataframe

        # Generate dict of dicts of dicts with shared perimeters according
        # to the requested adjacency rule
        adjacencies = neighbors(df, adjacency)
        graph = cls(adjacencies)

        graph.issue_warnings()

        # Add "exterior" perimeters to the boundary nodes
        add_boundary_perimeters(graph, df.geometry)

        # Add area data to the nodes
        areas = df.geometry.area.to_dict()
        networkx.set_node_attributes(graph, name="area", values=areas)

        return graph

    def add_data(self, df, columns=None):
        """Add columns of a DataFrame to a graph as node attributes using
        by matching the DataFrame's index to node ids.

        :param df: Dataframe containing given columns.
        :param columns: (optional) List of dataframe column names to add.
        """

        if columns is None:
            columns = df.columns

        check_dataframe(df[columns])

        column_dictionaries = df.to_dict("index")
        networkx.set_node_attributes(self, column_dictionaries)

    def node_attribute(self, node_attribute_key):
        """Create a dictionary of the form ``{node: <attribute value>}`` for
        the given attribute key, over all nodes of the graph.

        This is useful for creating an assignment dictionary from an attribute
        from a source data file. For example, if you created your graph from Census data
        and each node has a `CD` attribute that gives the congressional district
        the node belongs to, then `graph.node_attribute("CD")` would return the
        desired assignment of nodes to CDs.

        :param graph: NetworkX graph.
        :param node_attribute_key: Attribute available on all nodes.
        :return: Dictionary of {node_id: attribute} pairs.
        """
        return {node: data[node_attribute_key] for node, data in self.nodes.items()}

    def join(self, dataframe, columns=None, left_index=None, right_index=None):
        """Add data from a dataframe to the graph, matching nodes to rows when
        the node's `left_index` attribute equals the row's `right_index` value.

        :param dataframe: DataFrame.
        :columns: (optional) The columns whose data you wish to add to the graph.
            If not provided, all columns are added.
        :left_index: (optional) The node attribute used to match nodes to rows.
            If not provided, node IDs are used.
        :right_index: (optional) The DataFrame column name to use to match rows
            to nodes. If not provided, the DataFrame's index is used.
        """
        if right_index is not None:
            df = dataframe.set_index(right_index)
        else:
            df = dataframe

        if columns is not None:
            df = df[columns]

        check_dataframe(df)

        column_dictionaries = df.to_dict()

        if left_index is not None:
            ids_to_index = networkx.get_node_attributes(self, left_index)
        else:
            # When the left_index is node ID, the matching is just
            # a redundant {node: node} dictionary
            ids_to_index = dict(zip(self.nodes, self.nodes))

        node_attributes = {
            node_id: {
                column: values[index] for column, values in column_dictionaries.items()
            }
            for node_id, index in ids_to_index.items()
        }

        networkx.set_node_attributes(self, node_attributes)

    @property
    def islands(self):
        """The set of degree-0 nodes."""
        return set(node for node in self if self.degree[node] == 0)

    def warn_for_islands(self):
        """Issue a warning if the graph has any islands (degree-0 nodes)."""
        islands = self.islands
        if len(self.islands) > 0:
            warnings.warn(
                "Found islands (degree-0 nodes). Indices of islands: {}".format(islands)
            )

    def issue_warnings(self):
        """Issue warnings if the graph has any red flags (right now, only islands)."""
        self.warn_for_islands()


def add_boundary_perimeters(graph, geometries):
    """Add shared perimeter between nodes and the total geometry boundary.

    :param graph: NetworkX graph
    :param df: Geodataframe containing geometry information.
    :return: The updated graph.
    """
    prepared_boundary = prep(cascaded_union(geometries).boundary)

    boundary_nodes = geometries.boundary.apply(prepared_boundary.intersects)

    for node in graph:
        graph.nodes[node]["boundary_node"] = bool(boundary_nodes[node])
        if boundary_nodes[node]:
            total_perimeter = geometries[node].boundary.length
            shared_perimeter = sum(
                neighbor_data["shared_perim"] for neighbor_data in graph[node].values()
            )
            boundary_perimeter = total_perimeter - shared_perimeter
            graph.nodes[node]["boundary_perim"] = boundary_perimeter


def check_dataframe(df):
    for column in df.columns:
        if sum(df[column].isna()) > 0:
            warnings.warn("NA values found in column {}!".format(column))


def remove_geometries(data):
    """Remove geometry attributes from NetworkX adjacency data object,
    because they are not serializable. Mutates the ``data`` object.

    Does nothing if no geometry attributes are found.

    :param data: an adjacency data object (returned by
        :func:`networkx.readwrite.json_graph.adjacency_data`)
    """
    for node in data["nodes"]:
        bad_keys = []
        for key in node:
            # having a ``__geo_interface__``` property identifies the object
            # as being a ``shapely`` geometry object
            if hasattr(node[key], "__geo_interface__"):
                bad_keys.append(key)
        for key in bad_keys:
            del node[key]


def convert_geometries_to_geojson(data):
    """Convert geometry attributes in a NetworkX adjacency data object
    to GeoJSON, so that they can be serialized. Mutates the ``data`` object.

    Does nothing if no geometry attributes are found.

    :param data: an adjacency data object (returned by
        :func:`networkx.readwrite.json_graph.adjacency_data`)
    """
    for node in data["nodes"]:
        for key in node:
            # having a ``__geo_interface__``` property identifies the object
            # as being a ``shapely`` geometry object
            if hasattr(node[key], "__geo_interface__"):
                # The ``__geo_interface__`` property is essentially GeoJSON.
                # This is what :func:`geopandas.GeoSeries.to_json` uses under
                # the hood.
                node[key] = node[key].__geo_interface__



import random


def propose_any_node_flip(partition):
    """Propose to pick a random node to a
    random part

    """

    node = random.choice(tuple(partition.assignment.keys()))
    newpart = random.choice(partition.parts)

    return {node: newpart}


# def propose_random_flip_with_loops(partition):
#     """Proposes a random boundary flip from the partition.
#     Uses the number of cut edges to determine self--loops.

#     :param partition: The current partition to propose a flip from.
#     :return: a dictionary with the flipped node mapped to its new assignment

#     """
#     # self loop
#     numEdges = 2.0 * len(partition['cut_edges'])
#     if random.random() < 1.0 - (numEdges * 1.0 / max_edge_cuts(partition)):
#         return dict()

#     flip = propose_random_flip(partition)

#     # checks for a frozen nodes field and self loops if the value has
#     # been set to 1
#     flipped_node = list(flip.keys())[0]
#     node_attrs = partition.graph.nodes[flipped_node]
#     if "Frozen" in node_attrs and node_attrs["Frozen"]:
#         return dict()

#     return flip


# def propose_random_flip_metagraph(partition):
#     """Proposes a random boundary flip from the partition.
#     Uses the metagraph degree to determine self--loops.
#     Very slow.

#     :param partition: The current partition to propose a flip from.
#     :return: a dictionary with the flipped node mapped to its new assignment

#     """
#     # self loop
#     numEdges = partition["metagraph_degree"]
#     if random.random() < 1.0 - (numEdges * 1.0 / max_edge_cuts(partition)):
#         return dict()

#     flip = propose_random_flip(partition)

#     # checks for a frozen nodes field and self loops if the value has
#     # been set to 1
#     flipped_node = list(flip.keys())[0]
#     node_attrs = partition.graph.nodes[flipped_node]
#     if "Frozen" in node_attrs and node_attrs["Frozen"]:
#         return dict()

#     return flip


# def propose_several_random_flips(partition):
#     """Proposes between 2 and 7 random boundary flips from the partition.
#        Calls the propose_random_flip() method from this file.

#     :param partition: The current partition to propose a flip from.
#     :return: a dictionary with the flipped nodes mapped to their new assignments

#     """
#     number_of_flips = random.randint(2, 7)

#     proposal = dict()

#     for i in range(number_of_flips):
#         proposal.update(propose_random_flip(partition))

#     return proposal


def propose_flip_every_district(partition):
    """Proposes a random boundary flip for each district in the partition.

    :param partition: The current partition to propose a flip from.
    :return: a dictionary with the flipped nodes mapped to their new assignments

    """
    proposal = dict()

    for dist_edges in partition["cut_edges_by_part"].values():
        edge = random.choice(list(dist_edges))

        index = random.choice((0, 1))
        flipped_node, other_node = edge[index], edge[1 - index]
        flip = {flipped_node: partition.assignment[other_node]}

        proposal.update(flip)

    return proposal


def propose_chunk_flip(partition):
    """Chooses a random boundary node and proposes to flip it and all of its neighbors

    :param partition: The current partition to propose a flip from.
    :return: a dictionary with the flipped nodes mapped to their new assignments

    """
    proposal = dict()

    edge = random.choice(tuple(partition["cut_edges"]))
    index = random.choice((0, 1))

    flipped_node = edge[index]

    valid_flips = [
        nbr
        for nbr in partition.graph.neighbors(flipped_node)
        if partition.assignment[nbr] != partition.assignment[flipped_node]
    ]

    for flipped_neighbor in valid_flips:
        proposal.update({flipped_neighbor: partition.assignment[flipped_node]})

    return proposal


# def propose_single_or_chunk(partition):
#     """With probability .9, chooses a random boundary node and proposes to flip it.
#        With probability .1, chooses a random boundary node and proposes to flip it
#        and all of its neighbors.
#        Calls the propose_random_flip() and propose_chunk_flip()
#        methods from this file.


#     :param partition: The current partition to propose a flip from.
#     :return: a dictionary with the flipped nodes mapped to their new assignments

#     """
#     if(random.random() > .1):
#         return propose_random_flip(partition)
#     else:
#         return propose_chunk_flip(partition)


# def number_of_flips(partition, dict_of_flips, prev_partition):
#     flips = partition.flips
#     if flips is None or flips is prev_partition:
#         return dict_of_flips, prev_partition
#     else:
#         prev_partition = flips
#         dict_of_flips[next(iter(flips))] = dict_of_flips.get(next(iter(flips)), 0) + 1
#         return dict_of_flips, prev_partition


def propose_random_flip(partition):
    """Proposes a random boundary flip from the partition.

    :param partition: The current partition to propose a flip from.
    :return: a dictionary with the flipped node mapped to its new assignment

    """
    if len(partition["cut_edges"]) == 0:
        return dict()

    edge = random.choice(tuple(partition["cut_edges"]))
    index = random.choice((0, 1))

    flipped_node, other_node = edge[index], edge[1 - index]

    flip = {flipped_node: partition.assignment[other_node]}

    return flip


# def propose_lowest_pop_single_flip(partition):
#     dists = list(partition['population'])
#     pops = partition['population'].values()
#     rev_pop = dict(zip(pops, dists))
#     dist = rev_pop[min(pops)]

#     edge = random.choice(tuple(partition['cut_edges_by_part'][dist]))

#     if partition.assignment[edge[0]] == dist:
#         flip = {edge[1]: dist}
#     else:
#         flip = {edge[0]: dist}

#     return flip


# def propose_single_lowest_pop_or_random(partition):
#     if(random.random() > .2):
#         return propose_random_flip(partition)
#     else:
#         return propose_lowest_pop_single_flip(partition)


# def propose_chunk_swap(partition):
#     proposal = dict()
#     dists = list(partition.parts)

#     while len(dists) != 0:
#         dist = dists[0]

#         edge = random.choice(list(partition['cut_edges_by_part'][dist]))
#         index = random.choice((0, 1))
#         flipped_node = edge[index]

#         if partition.assignment[flipped_node] != dist:
#             proposal[flipped_node] = dist

#         valid_flips = []
#         count = 0

#         for nbr in partition.graph.neighbors(flipped_node):
#             if len(valid_flips) == 0:
#                 if partition.assignment[nbr] != dist:
#                     valid_flips.append(nbr)
#                     count = count + 1

#             else:
#                 a = partition.assignment[valid_flips[0]]
#                 if a == partition.assignment[nbr]:
#                     valid_flips.append(nbr)
#                     count = count + 1

#         for flipped_neighbor in valid_flips:
#             proposal[flipped_neighbor] = dist

#         dists.remove(dist)

#     return proposal


# def reversible_chunk_flip(partition):
#     edge = random.choice(tuple(partition['cut_edges']))
#     index = random.choice((0, 1))

#     flipped_node, other_node = edge[index], edge[1 - index]
#     flip_to = partition.assignment[flipped_node]
#     flip_from = partition.assignment[other_node]

#     num_flips = 1
#     flips = [flipped_node]
#     choices = [nbr for nbr in partition.graph.neighbors(flipped_node)
#                if partition.assignment[nbr] == flip_from]
#     while(choices and random.random() < .5 ** num_flips):
#         next_flip = random.choice(tuple(choices))
#         flips.append(next_flip)
#         num_flips += 1
#         choices.remove(next_flip)
#         new_choices = [nbr for nbr in partition.graph.neighbors(next_flip)
#                        if partition.assignment[nbr] == flip_from and nbr not in flips]
#         choices = list(set(choices) | set(new_choices))
#     return {flip: flip_to for flip in flips}


def max_edge_cuts(partition):
    # """returns wes computation for max number of edge cuts... not well documented,
    # and a vague upper bound (to be made smaller if possible)

    # inputs:
    # :param partition: a partition instance.

    # returns: an integer value

    # """
    # TODO need number of frozen edges of graph
    numFrozen = 0
    numDists = len(partition)
    return 2 * (2 * len(partition.graph.nodes) + (numDists - numFrozen) - 6)



from gerrychain.accept import always_accept
from gerrychain.chain import MarkovChain
from gerrychain.constraints import (
    Validator,
    no_vanishing_districts,
    single_flip_contiguous,
)
from gerrychain.proposals import propose_random_flip

default_constraints = [single_flip_contiguous]


class DefaultChain(MarkovChain):
    """
    A MarkovChain with propose_random_flips proposal and always_accept
    acceptance function. Also instantiates a Validator for you from a
    list of constraints.
    """

    def __init__(self, partition, constraints, total_steps):
        if len(constraints) > 0:
            validator = Validator(constraints)
        else:
            validator = lambda x: True
        super().__init__(
            propose_random_flip, validator, always_accept, partition, total_steps
        )


grid_validator = Validator([single_flip_contiguous, no_vanishing_districts])


class GridChain(MarkovChain):
    """
    A basic MarkovChain for use with the Grid partition. The proposal is a single random flip
    at the boundary of a district. A step is valid if the districts are connected and no
    districts disappear. Requires a 'cut_edges' updater.
    """

    def __init__(self, initial_grid, total_steps=1000):
        if not initial_grid["cut_edges"]:
            raise ValueError(
                "BasicChain needs the Partition to have a cut_edges updater."
            )

        super().__init__(
            propose_random_flip,
            grid_validator,
            always_accept,
            initial_grid,
            total_steps=total_steps,
        )



import json
import math
from collections import Counter

from gerrychain.scores import efficiency_gap, mean_median, mean_thirdian


def p_value_report(score_name, ensemble_scores, initial_plan_score):
    """
    p_value_report takes an iterable :ensemble_scores: (scores computed on the
    ensemble seen in the random walk) and the :initial_plan_score: and returns
    a report like page 5 of Moon's PA analysis, including the fraction of plans
    in the ensemble that scored higher than the initial score and the p-value
    that the plan is chosen randomly as determined by the 'square root of
    2*epsilon' theorem.
    """
    count_higher = Counter(score >= initial_plan_score for score in ensemble_scores)
    total_scores = sum(count_higher.values())
    fraction_as_high = count_higher[True] / total_scores
    fraction_lower = count_higher[False] / total_scores

    # By Chikina-Frieze-Pegden, if a plan scores in the highest fraction_higher of all
    # plans seen in a random walk, then it has less than sqrt(2 * fraction_higher)
    # probability of being chosen randomly from the sample space. (Paraphrasing Moon's
    # PA report).

    p_value = math.sqrt(2 * fraction_as_high)
    opposite_p_value = math.sqrt(2 * fraction_lower)

    report = {'name': score_name,
              'initial_plan_score': initial_plan_score,
              'fraction_as_high': fraction_as_high,
              'p_value': p_value,
              'opposite_p_value': opposite_p_value}
    return report


class SlimPValueReport:
    scores = {
        "Efficiency Gap": efficiency_gap,
        "Mean-Median": mean_median,
        "Mean-Thirdian": mean_thirdian
    }

    def __init__(self, election, function=lambda epsilon: math.sqrt(2 * epsilon), scores=None):
        if scores is not None:
            self.scores = scores

        self.election = election
        self.function = function

        self.counters = {name: Counter() for name in self.scores}
        self.histograms = {name: Histogram((0, 1), 10000) for name in self.scores}
        self.initial_scores = None

    def __call__(self, partition):
        election_results = partition[self.election.alias]

        if self.initial_scores is None:
            self.initial_scores = {name: score(election_results)
                                   for name, score in self.scores.items()}

        for name, score in self.scores.items():
            value = score(election_results)
            self.counters[name].update([value >= self.initial_scores[name]])
            self.histograms[name].record(value)

    def render(self):
        output = self.render_without_histograms()
        for record in output["analysis"]:
            record["histogram"] = self.render_histogram(record["score"])
        return output

    def render_without_histograms(self):
        return {"election": self.election.name,
                "analysis": [self.render_score_analysis(score)
                            for score in self.scores]}

    def compute_p_value(self, score_name):
        return self.function(self.fraction_as_high(score_name))

    def compute_opposite_p_value(self, score_name):
        return self.function(self.fraction_as_low(score_name))

    def fraction_as_high(self, score_name):
        size_of_ensemble = sum(self.counters[score_name].values())
        if size_of_ensemble == 0:
            return math.nan
        number_as_high = self.counters[score_name][True]
        return number_as_high / size_of_ensemble

    def fraction_as_low(self, score_name):
        size_of_ensemble = sum(self.counters[score_name].values())
        if size_of_ensemble == 0:
            return math.nan
        number_as_low = 1 + self.counters[score_name][False]
        return number_as_low / size_of_ensemble

    def render_score_analysis(self, score):
        first_party, *others = self.election.parties_to_columns
        party_name = format_party_name(first_party)
        return {"score": score,
                "interpretation": "Positive values mean an advantage for {}".format(party_name),
                "fraction_as_high_as_original": self.fraction_as_high(score),
                "number_as_high_as_original": self.counters[score][True],
                "number_lower_than_original": self.counters[score][False],
                "p": self.compute_p_value(score)}

    def render_histogram(self, score):
        return self.histograms[score].json()

    def __str__(self):
        return json.dumps(self.render(), indent=2)


def format_party_name(name):
    if name.lower()[:-5] == "party":
        return name
    return "{} Party".format(name)


class Histogram:
    """
    A histogram with fixed bins, determined by the number of bins and the
    bounds for values.

    Anthony is working on the super smart histogram that does the binning for you.
    """

    def __init__(self, bounds, number_of_bins):
        self.bounds = bounds
        self.number_of_bins = number_of_bins

        left, right = bounds
        self.bin_size = abs(right - left) / number_of_bins

        self.counter = Counter()

    def record(self, value):
        """
        :value: value to record in the histogram
        """
        return self.counter.update([self.find_bin_index(value)])

    def find_bin_index(self, value):
        left = self.bounds[0]
        bin_index = math.floor((value - left) / self.bin_size)
        return bin_index

    def csv(self):
        header = "left,right,count\n"
        rows = ",".join((left, right, count) for (left, right), count in self.json())
        body = "\n".join(rows)
        return header + body

    def json(self):
        def get_bin(bin_index):
            return (self.bin_size * bin_index, self.bin_size * (bin_index + 1))
        return [(get_bin(index), count) for index, count in self.counter.items()]


class ChainOutputTable:
    def __init__(self, data=None):
        if not data:
            data = list()
        self.data = data

    def append(self, row):
        self.data.append(row)

    def to_json(self, filename=None):
        if filename is None:
            return json.dumps(self.data)

        with open(filename, "w") as f:
            json.dump(self.data, f)

    # def to_dataframe(self):
    #     return pd.DataFrame(self.data)

    def __iter__(self):
        return self

    def __next__(self):
        return self.data

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.data[key]
        else:
            return [row[key] for row in self.data]

    def district(self, district_id):
        return get_from_each(self.data, district_id)


def get_chain_scores(chain, handlers):
    """Yield scores from handlers on each state in the chain.

    :param chain: :class:`gerrychain.chain.MarkovChain` instance.
    :param handlers: Dictionary of {name: func} pairs.
    :return: A generator yielding dictionaries of {name: result} pairs from
              handlers.

    """
    return ({key: handler(state) for key, handler in handlers.items()}
            for state in chain)


def handle_scores_separately(chain, handlers):
    table = ChainOutputTable()

    initialScores = {key: handler(chain.state) for
            key, handler in handlers.items() if key != "flips"}
    table.append(initialScores)

    nhandlers = {key: value for key, value in handlers.items() if key != "flips"}

    jsonToText = '{'
    jsonSave = False
    if "flips" in list(handlers.keys()):
        jsonToText += '"0": ' + json.dumps(handlers['flips'](chain.state))
        jsonSave = True

    for row in get_chain_scores(chain, nhandlers):
        table.append(row)
        if jsonSave:
            jsonToText += ", " + "\"" + str(chain.counter + 1) + "\""
            + ": " + json.dumps(handlers["flips"](chain.state))
    jsonToText += '}'

    return (table, jsonToText, nhandlers)


def get_from_each(table, key):
    return [{header: row[header][key] for header in row if key in row[header]}
            for row in table]


def log_dict_as_json(hist, scores, outputFile="output.json"):
    print(outputFile)
    if outputFile is not None:
        with open(outputFile, "w") as f:
            f.write(json.dumps(hist))


def flips_to_pngs(hist, scores, outputFile="output.png"):
    pass


def log_table_to_file(table, scores, outputFile="output.txt"):
    if outputFile:
        with open(outputFile, 'w') as f:
            for key in scores:
                if table[key] is not None:
                    f.write(", ".join([str(x) for x in table[key]]))
                    f.write("\n")


def pipe_to_table(chain, handlers, display=True, number_to_display=10,
                  bin_interval=1):
    table = ChainOutputTable()

    if number_to_display > 0:
        display_interval = math.floor(len(chain) / number_to_display)
    else:
        display_interval = 1

    counter = 0
    for row in handle_chain(chain, handlers):
        if display and counter % display_interval == 0:
            print("Step {}".format(counter))
            print(row)
        if counter % bin_interval == 0:
            table.append(row)
        counter += 1
    return table


def handle_chain(chain, handlers):
    for state in chain:
        yield {key: handler(state) for key, handler in handlers.items()}



from collections import Counter

from gerrychain.vendor.utm import from_latlon


def utm_of_point(point):
    return from_latlon(point.y, point.x)[2]


def identify_utm_zone(df):
    wgs_df = df.to_crs("+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs")
    utm_counts = Counter(utm_of_point(point) for point in wgs_df["geometry"].centroid)
    # most_common returns a list of tuples, and we want the 0,0th entry
    most_common = utm_counts.most_common(1)[0][0]
    return most_common


def reprojected(df):
    """Returns a copy of `df`, projected into the coordinate reference system of a suitable
        `Universal Transverse Mercator`_ zone.
    :param df: :class:`geopandas.GeoDataFrame`
    :rtype: :class:`geopandas.GeoDataFrame`

    .. _`Universal Transverse Mercator`: https://en.wikipedia.org/wiki/UTM_coordinate_system
    """
    utm = identify_utm_zone(df)
    return df.to_crs(
        f"+proj=utm +zone={utm} +ellps=WGS84 +datum=WGS84 +units=m +no_defs"
    )



from ..updaters.flows import flows_from_changes
from ..utils import level_sets


class Assignment:
    """An assignment of nodes into parts.

    The goal of :class:`Assignment` is to provide an interface that mirrors a
    dictionary (what we have been using for assigning nodes to districts) while making it
    convenient/cheap to access the set of nodes in each part.

    An :class:`Assignment` has a ``parts`` property that is a dictionary of the form
    ``{part: <frozenset of nodes in part>}``.
    """

    def __init__(self, parts: dict):
        self.parts = parts

    @classmethod
    def from_dict(cls, assignment: dict):
        """Create an Assignment from a dictionary. This is probably the method you want
        to use to create a new assignment.
        """
        parts = {
            part: frozenset(nodes) for part, nodes in level_sets(assignment).items()
        }
        return cls(parts)

    def __getitem__(self, node):
        for part, nodes in self.parts.items():
            if node in nodes:
                return part
        raise KeyError(node)

    def copy(self):
        """Returns a copy of the assignment.
        Does not duplicate the frozensets of nodes, just the parts dictionary.
        """
        return Assignment(self.parts.copy())

    def update(self, mapping: dict):
        """Update the assignment for some nodes using the given mapping.
        """
        flows = flows_from_changes(self, mapping)
        for part, flow in flows.items():
            # Union between frozenset and set returns an object whose type
            # matches the object on the left, which here is a frozenset
            self.parts[part] = (self.parts[part] - flow["out"]) | flow["in"]

    def items(self):
        """Iterate over ``(node, part)`` tuples, where ``node`` is assigned to ``part``.
        """
        for part, nodes in self.parts.items():
            for node in nodes:
                yield (node, part)

    def update_parts(self, new_parts):
        """Update some parts of the assignment. Does not check that every node is
        still assigned to a part.

        :param dict new_parts: dictionary mapping (some) parts to their new sets or
            frozensets of nodes
        """
        for part, nodes in new_parts.items():
            self.parts[part] = frozenset(nodes)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __call__(self, key):
        return self[key]


def get_assignment(assignment, graph=None):
    if isinstance(assignment, str):
        if graph is None:
            raise TypeError(
                "You must provide a graph when using a node attribute for the assignment"
            )
        return Assignment.from_dict(graph.node_attribute(assignment))
    elif isinstance(assignment, dict):
        return Assignment.from_dict(assignment)
    elif isinstance(assignment, Assignment):
        return assignment
    else:
        raise TypeError("Assignment must be a dict or a node attribute key")
